open Tree
open Absyntax

module type CANON = 
sig
    val linearize : Tree.stm -> Tree.stm list
        (* From an arbitrary Tree statement, produce a list of cleaned trees
	   satisfying the following properties:
	      1.  No SEQ's or ESEQ's
	      2.  The parent of every CALL is an EXP(..) or a MOVE(TEMP t,..)
        *)

    val basicBlocks : Tree.stm list -> (Tree.stm list list * Tree.label)
        (* From a list of cleaned trees, produce a list of
	 basic blocks satisfying the following properties:
	      1. and 2. as above;
	      3.  Every block begins with a LABEL;
              4.  A LABEL appears only at the beginning of a block;
              5.  Any JUMP or CJUMP is the last stm in a block;
              6.  Every block ends with a JUMP or CJUMP;
           Also produce the "label" to which control will be passed
           upon exit.
        *)

    val traceSchedule : Tree.stm list list * Tree.label -> Tree.stm list
         (* From a list of basic blocks satisfying properties 1-6,
            along with an "exit" label,
	    produce a list of stms such that:
	      1. and 2. as above;
              7. Every CJUMP(_,t,f) is immediately followed by LABEL f.
            The blocks are reordered to satisfy property 7; also
	    in this reordering as many JUMP(T.NAME(lab)) statements
            as possible are eliminated by falling through into T.LABEL(lab).
         *)
end

module Canon (* : CANON *) = 
struct

 module T = Tree

(* let linearize (stm0: T.stm) : T.stm list = *)
 let (%) a b = 
   match (a,b) 
   with
     ((T.EXP(T.CONST _)),x) -> x
   | (x, (T.EXP(T.CONST _))) -> x        
   | (x,y) -> T.SEQ(x,y)

 and commute = function
      (T.EXP(T.CONST _), _) -> true
    | (_, T.NAME _) -> true
    | (_, T.CONST _) -> true
    |  _ -> false

 and nop = T.EXP(T.CONST 0)

 (* in *)
let rec reorder = function
      ((T.CALL _ as e)::rest) ->
	let t = Temp.new_temp()
	 in reorder(T.ESEQ(T.MOVE(T.TEMP t, e), T.TEMP t) :: rest)
    | (a::rest) ->
	 let (stms,e) = do_exp a
	 and (stms',el) = reorder rest
	 in if commute(stms',e)
	    then (stms % stms',e::el)
	    else let t = Temp.new_temp()
		   in (stms % T.MOVE(T.TEMP t, e) % stms', T.TEMP t :: el)
    | [] -> (nop,[])

  and reorder_exp (el,build) = let (stms,el') = reorder el
                               in (stms, build el')

  and reorder_stm (el,build) = let (stms,el') = reorder (el)
		 	       in stms % build(el')

  and do_stm = function
      (T.SEQ(a,b)) -> do_stm a % do_stm b
    | (T.JUMP(e,labs)) -> reorder_stm([e],fun [e] -> T.JUMP(e,labs))
    | (T.CJUMP(p,a,b,t,f)) -> reorder_stm([a;b], fun[a;b]-> T.CJUMP(p,a,b,t,f))
    | (T.MOVE(T.TEMP t,T.CALL(e,el))) -> 
         reorder_stm(e::el,fun (e::el) -> T.MOVE(T.TEMP t,T.CALL(e,el)))
    | (T.MOVE(T.TEMP t,b)) -> reorder_stm([b],fun[b]->T.MOVE(T.TEMP t,b))
    | (T.MOVE(T.MEM e,b)) -> reorder_stm([e;b],fun[e;b]->T.MOVE(T.MEM e,b))
    | (T.MOVE(T.ESEQ(s,e),b)) -> do_stm(T.SEQ(s,T.MOVE(e,b)))
    | (T.EXP(T.CALL(e,el))) -> reorder_stm(e::el,fun (e::el) -> T.EXP(T.CALL(e,el)))
    | (T.EXP e) -> reorder_stm([e],fun[e]->T.EXP e)
    |  s -> reorder_stm([],fun[]->s)

  and do_exp = function 
      (T.BINOP(p,a,b)) -> reorder_exp([a;b], fun[a;b]->T.BINOP(p,a,b))
    | (T.MEM(a)) -> reorder_exp([a], fun[a]->T.MEM(a))
    | (T.ESEQ(s,e)) -> 
           let stms = do_stm s
           and  (stms',e) = do_exp e
            in (stms%stms',e)
    | (T.CALL(e,el) as ex ) -> reorder_exp([ex], fun [ex'] -> ex')
    |  e -> reorder_exp([],fun[]->e)

  (* linear gets rid of the top-level SEQ's, producing a list *)
(*  in *)
let rec linear = function 
      (T.SEQ(a,b),l) -> linear(a,linear(b,l))
    | (s,l) -> s::l

(* in *)(* body of linearize *)
 let linearize (stm0: T.stm) : T.stm list = (* *)
    linear(do_stm stm0, [])

  type block = T.stm list

  (* Take list of statements and make basic blocks satisfying conditions
       3 and 4 above, in addition to the extra condition that 
      every block ends with a JUMP or CJUMP *)

  let basicBlocks stms = 
     let donel = Temp.new_label "done" in 
     let rec blocks = 
           function
               ((T.LABEL _ as head) :: tail, blist) ->
	         let rec next = function 
                      (((T.JUMP _) as s)::rest, thisblock) -> endblock(rest, s::thisblock)
		   | (((T.CJUMP _)as s )::rest, thisblock) ->
                            endblock(rest,s::thisblock)
		   | ((T.LABEL lab :: _) as stms, thisblock) ->
                            next(T.JUMP(T.NAME lab,[lab]) :: stms, thisblock)
		   | (s::rest, thisblock) -> next(rest, s::thisblock)
		   | ([], thisblock) ->
			     next([T.JUMP(T.NAME donel, [donel])], thisblock)
		 and endblock(stms, thisblock) = blocks(stms, List.rev thisblock :: blist)
	     in next(tail, [head])
	   | ([], blist) -> List.rev blist
	   | (stms, blist) -> blocks(T.LABEL(Temp.new_label "block_entry")::stms, blist)
      in (blocks(stms,[]), donel)

  let enterblock = function
      ((T.LABEL s :: _ as b), table) -> Symbol.add s b table
    | (_, table) -> table

  let rec splitlast = function ([x]) -> ([],x)
    | (h::t) -> let (t',last) = splitlast t in (h::t', last)

  let rec trace(table,((T.LABEL lab :: _) as b),rest) = 
    let table = Symbol.add lab [] table
    in match splitlast b
       with (most,T.JUMP(T.NAME lab, _)) ->
	  (match Symbol.look lab table
            with Some(_::_ as b') -> most @ trace(table, b', rest)
	     | _ -> b @ getnext(table,rest))
       | (most,T.CJUMP(opr,x,y,t,f)) ->
          (match (Symbol.look t table, Symbol.look f table)
           with (_, Some((_::_) as b')) -> b @ trace(table, b', rest)
             | (Some((_::_) as b'), _) -> 
		           most @ [T.CJUMP(T.notRel opr,x,y,f,t)]
		                @ trace(table, b', rest)
             | _ -> let f' = Temp.new_label "trace"
		     in most @ [T.CJUMP(opr,x,y,t,f');
				T.LABEL f'; T.JUMP(T.NAME f,[f])]
			     @ getnext(table,rest)
                        )
      | (most, T.JUMP _) -> b @ getnext(table,rest)

  and getnext = function
      (table,((T.LABEL lab::_) as b)::rest) ->
           (match Symbol.look lab table
            with Some(_::_) -> trace(table,b,rest)
              | _ -> getnext(table,rest))
    | (table,[]) -> []

  let rec simplify =
    function
        [] -> []
      | (T.JUMP (T.NAME l,ll) as a )::(T.LABEL l' as b)::r -> 
                  if l=l' then simplify r else a::b::(simplify r)
      |  a::r -> a::(simplify r)

  let traceSchedule(blocks,donel) = 
       simplify(getnext(List.fold_right (fun x -> fun y -> enterblock (x,y)) blocks (Symbol.mkempty()), blocks)
         @ [T.LABEL donel])
 end
