open Absyntax;;
open Absyn;;
open Symbol;;

(* Module Types: la representation des types utilises dans le type-checker *)

module Types =
struct
  type unique = unit ref
  type ty = 
            RECORD of (Symbol.symbol * ty) list * unique
          | NIL
          | INT
          | BOOL
          | CHAR
          | FLOAT
          | STRING
          | ARRAY of ty * unique
          | NAME of Symbol.symbol * ty option ref
          | UNIT

  module Assl = Map.Make(struct type t = ty let compare = compare end)
end


open Types;;


(* The definition of the type and value Environments we will use in the type checker code ... *)

module Env =
struct
  type ty = Types.ty
  type enventry = VarEntry of ty | FunEntry of ty list * ty (* types des parametres et resultat (UNIT si procedure) *)
  let base_tenv = maketable (List.map (fun (n,v) -> (Symbol.symbol n,v))
                  [
                   ("int",INT);
                   ("float",FLOAT);
                   ("char",CHAR);
                   ("string",STRING);
                  ])
  let base_venv = maketable (List.map (fun (n,v) -> (Symbol.symbol n,v))
                  [
                   ("printint",FunEntry([INT],UNIT));
                   ("ord",FunEntry([CHAR],INT));
                   ("chr",FunEntry([INT],CHAR));
                   ("size",FunEntry([STRING],INT));
                   ("print",FunEntry([STRING],UNIT));
                   ("concat",FunEntry([STRING;STRING],STRING));
                   ("noop",VarEntry(UNIT));
                   ("getchar",FunEntry([],CHAR)); 
                   ("readint",FunEntry([],INT)); 
                   ("getstring",FunEntry([INT],STRING));
                   ("mkstring",FunEntry([CHAR],STRING))
                  ])
end

open Env;; (* we will use it a lot *)


(* Fonctions auxiliaires pour le type checker *)

(* Traduction simpliste de Absyn.coretype vers Types.ty *)
(*                                                      *)
(* let rec transTy tenv =
  function
      {typ_desc=Typ_name(s)} -> (find s tenv)
    | {typ_desc=Typ_array(t)} -> ARRAY(find t tenv,ref ())
    | {typ_desc=Typ_record(l)} -> RECORD(List.map (fun {f_name=n;typ=t} 
                                                       -> (n, find t tenv)
                                                  ) l,ref ())
;;
*)

(* suivre une chaine de definition de types *)
let rec actual_ty =
  function
      NAME(s,{contents=Some(t)}) -> actual_ty t
    |  t                     -> t;;

(* Traduction d'un groupe de declarations recursives (symbol, Absyn.coretype) vers une liste de (symbol, Types.ty) *)

let rec transTyRecDecl tenv =
    function tdecls -> (* Ajoutons d'abord dans l'environnement les noms des types *)
                       let tenv' = 
                         (List.fold_left (fun tenv -> 
                                          fun (n,tdef) -> add n (NAME(n,ref None)) tenv) tenv tdecls) 
                       in (* Maintenant, pour chaque "corps" de definition, on calcule le
                             type et on l'affecte a la reference qui se trouve dans l'environnement *)
                          List.map 
                          (fun (n,tdef) 
                               -> (n,
                                   let t = transTyRec tenv' tdef in
                                   match (find n tenv') with
				     (NAME(n,r)) -> r := Some t; t
				   | _ -> t
                          )) tdecls

and transTyRec tenv =  (* ici, nous creons les boucles qu'il faut *)
    function  
      {typ_desc=Typ_name(s)} -> skipnames (find s tenv) (* pointe sur le NAME(n,None) 
                                                           ou sur le t du NAME(n,Some t) de l'env *)
    | {typ_desc=Typ_array(t)} -> ARRAY(find t tenv,ref ())
    | {typ_desc=Typ_record(l)} -> RECORD(List.map (fun {f_name=n;typ=t} 
                                                       -> (n,find t tenv)
                                                  ) l,ref ())
and skipnames = function (* compression des parcours *)
        NAME(s,{contents=None}) as t -> t
  |     NAME(s,{contents=Some t}) -> skipnames t
  |     t -> t
;;

(* test cases pour ransTyRecDecl *)

let decls = [(Symbol.symbol "a"), {typ_desc=Typ_name (Symbol.symbol "int"); typ_loc=9, 12};
 (Symbol.symbol "b"), {typ_desc=Typ_name (Symbol.symbol "a"); typ_loc=22, 23};
 (Symbol.symbol "c"),
 {typ_desc=Typ_array (Symbol.symbol "b"); typ_loc=33, 43}]
;;

let recdecls = [(Symbol.symbol "a"), {typ_desc=Typ_array (Symbol.symbol "c"); typ_loc=9, 12};
                (Symbol.symbol "b"), {typ_desc=Typ_name (Symbol.symbol "a"); typ_loc=22, 23};
                (Symbol.symbol "c"), {typ_desc=Typ_array (Symbol.symbol "b"); typ_loc=33, 43}]
;;

let testnonrec () = transTyRecDecl (add (Symbol.symbol "int") INT (mkempty ())) decls;;
let testrec () = transTyRecDecl (add (Symbol.symbol "int") INT (mkempty ())) recdecls;;


